@@ -130,8 +130,9 @@ is_flush_sig (int sig)
 }
 
 void
-tty_min::kill_pgrp (int sig)
+tty_min::kill_pgrp (int sig, pid_t target_pgid)
 {
+  target_pgid = target_pgid ?: pgid;
   bool killself = false;
   if (is_flush_sig (sig) && cygheap->ctty)
     cygheap->ctty->sigflush ();
@@ -145,8 +146,11 @@ tty_min::kill_pgrp (int sig)
   for (unsigned i = 0; i < pids.npids; i++)
     {
       _pinfo *p = pids[i];
-      if (!p || !p->exists () || p->ctty != ntty || p->pgid != pgid)
+      if (!p || !p->exists () || p->ctty != ntty || p->pgid != target_pgid)
 	continue;
+      if (p->process_state & PID_NOTCYGWIN)
+	continue; /* Do not send signal to non-cygwin process to prevent
+		     cmd.exe from crash. */
       if (p == myself)
 	killself = sig != __SIGSETPGRP && !exit_state;
       else
@@ -306,6 +310,222 @@ fhandler_termios::echo_erase (int force)
     doecho ("\b \b", 3);
 }
 
+/* The basic policy is as follows:
+   - The signal generated by key press will be sent only to cygwin process.
+   - For non-cygwin process, CTRL_C_EVENT will be sent on Ctrl-C. */
+/* The inferior of GDB is an exception. GDB does not support to hook signal
+   even if the inferior is a cygwin app. As a result, inferior cannot be
+   continued after interruption by Ctrl-C if SIGINT was sent. Therefore,
+   CTRL_C_EVENT rather than SIGINT is sent to the inferior of GDB. */
+fhandler_termios::process_sig_state
+fhandler_termios::process_sigs (char c, tty* ttyp, fhandler_termios *fh)
+{
+  termios &ti = ttyp->ti;
+  pid_t pgid = ttyp->pgid;
+
+  /* The name *_nat stands for 'native' which means non-cygwin apps. */
+  bool ctrl_c_event_sent = false;
+  bool need_discard_input = false;
+  bool pg_with_nat = false; /* The process group has non-cygwin processes. */
+  bool need_send_sig = false; /* There is process which need the signal. */
+  bool nat_shell = false; /* The shell seems to be a non-cygwin process. */
+  bool cyg_reader = false; /* Cygwin process is reading the tty. */
+  bool with_debugger = false; /* GDB is debugging cygwin app. */
+  bool with_debugger_nat = false; /* GDB is debugging non-cygwin app. */
+
+  winpids pids ((DWORD) 0);
+  for (unsigned i = 0; i < pids.npids; i++)
+    {
+      _pinfo *p = pids[i];
+      /* PID_NOTCYGWIN: check this for non-cygwin process.
+	 PID_NEW_PG: check this ofr GDB with non-cygwin inferior in pty
+		     without pcon enabled. In this case, the inferior is not
+		     cygwin process list. PID_NEW_PG is set as a marker for
+		     GDB with non-cygwin inferior in pty code.
+	 !PID_CYGPARENT: check this for GDB with cygwin inferior or
+			 cygwin apps started from non-cygwin shell. */
+      if (c == '\003' && p && p->ctty == ttyp->ntty && p->pgid == pgid
+	  && ((p->process_state & PID_NOTCYGWIN)
+	      || ((p->process_state & PID_NEW_PG)
+		  && ttyp->pcon_input_state_eq (tty::to_nat))
+	      || !(p->process_state & PID_CYGPARENT)))
+	{
+	  /* Ctrl-C event will be sent only to the processes attaching
+	     to the same console. Therefore, attach to the console to
+	     which the target process is attaching before sending the
+	     CTRL_C_EVENT. After sending the event, reattach to the
+	     console to which the process was previously attached.  */
+	  bool console_exists = fhandler_console::exists ();
+	  acquire_attach_mutex (mutex_timeout);
+	  pinfo pinfo_resume = pinfo (myself->ppid);
+	  DWORD resume_pid = 0;
+	  if (pinfo_resume)
+	    resume_pid = pinfo_resume->dwProcessId;
+	  else
+	    resume_pid = fhandler_pty_common::get_console_process_id
+	      (myself->dwProcessId, false);
+	  if ((!console_exists || resume_pid) && fh && !fh->is_console ())
+	    {
+	      FreeConsole ();
+	      AttachConsole (p->dwProcessId);
+	      init_console_handler (false);
+	    }
+	  if (fh && p == myself && being_debugged ())
+	    { /* Avoid deadlock in gdb on console. */
+	      fh->tcflush(TCIFLUSH);
+	      fh->release_input_mutex_if_necessary ();
+	    }
+	  /* CTRL_C_EVENT does not work for the process started with
+	     CREATE_NEW_PROCESS_GROUP flag, so send CTRL_BREAK_EVENT
+	     instead. */
+	  if ((p->process_state & PID_NEW_PG)
+	      && (p->process_state & PID_NOTCYGWIN))
+	    GenerateConsoleCtrlEvent (CTRL_BREAK_EVENT, p->dwProcessId);
+	  else if ((!fh || fh->need_send_ctrl_c_event ()
+		    || p->exec_dwProcessId == p->dwProcessId)
+		   && !ctrl_c_event_sent)
+	    {
+	      GenerateConsoleCtrlEvent (CTRL_C_EVENT, 0);
+	      ctrl_c_event_sent = true;
+	    }
+	  if ((!console_exists || resume_pid) && fh && !fh->is_console ())
+	    {
+	      /* If a process on pseudo console is killed by Ctrl-C,
+		 this process may take over the ownership of the
+		 pseudo console because this process attached to it
+		 before sending CTRL_C_EVENT. In this case, closing
+		 pseudo console is necessary. */
+	      fhandler_pty_slave::close_pseudoconsole_if_necessary (ttyp, fh);
+	      FreeConsole ();
+	      if (resume_pid && console_exists)
+		AttachConsole (resume_pid);
+	      init_console_handler (false);
+	    }
+	  release_attach_mutex ();
+	  need_discard_input = true;
+	}
+      if (p && p->ctty == ttyp->ntty && p->pgid == pgid)
+	{
+	  if (p->process_state & PID_NOTCYGWIN)
+	    pg_with_nat = true; /* The process group has non-cygwin process */
+	  if (!(p->process_state & PID_NOTCYGWIN))
+	    need_send_sig = true; /* Process which needs signal exists */
+	  if (!p->cygstarted)
+	    nat_shell = true; /* The shell seems to a non-cygwin shell */
+	  if (p->process_state & PID_TTYIN)
+	    cyg_reader = true; /* Theh process is reading the tty */
+	  if (!p->cygstarted && !(p->process_state & PID_NOTCYGWIN)
+	      && (p->process_state & PID_DEBUGGED))
+	    with_debugger = true; /* inferior is cygwin app */
+	  if (!(p->process_state & PID_NOTCYGWIN)
+	      && (p->process_state & PID_NEW_PG) /* Check marker */
+	      && ttyp->pcon_input_state_eq (tty::to_nat)
+	      && p->pid == pgid)
+	    with_debugger_nat = true; /* inferior is non-cygwin app */
+	}
+    }
+  if ((with_debugger || with_debugger_nat) && need_discard_input)
+    {
+      if (!(ti.c_lflag & NOFLSH) && fh)
+	{
+	  fh->eat_readahead (-1);
+	  fh->discard_input ();
+	}
+      ti.c_lflag &= ~FLUSHO;
+      return done_with_debugger;
+    }
+  if (with_debugger_nat)
+    return not_signalled; /* Do not process slgnal keys further.
+			     The non-cygwin inferior of GDB cannot receive
+			     the signals. */
+  /* Send SIGQUIT to non-cygwin process. Non-cygwin app will not be alerted
+     by kill_pgrp(), however, QUIT key should quit the non-cygwin app
+     if it is started along with cygwin process from cygwin shell. */
+  if ((ti.c_lflag & ISIG) && CCEQ (ti.c_cc[VQUIT], c)
+      && pg_with_nat && need_send_sig && !nat_shell)
+    {
+      for (unsigned i = 0; i < pids.npids; i++)
+	{
+	  _pinfo *p = pids[i];
+	  if (p && p->ctty == ttyp->ntty && p->pgid == pgid
+	      && (p->process_state & PID_NOTCYGWIN))
+	    sig_send (p, SIGQUIT);
+	}
+    }
+  if ((ti.c_lflag & ISIG) && need_send_sig)
+    {
+      int sig;
+      if (CCEQ (ti.c_cc[VINTR], c))
+	sig = SIGINT;
+      else if (CCEQ (ti.c_cc[VQUIT], c))
+	sig = SIGQUIT;
+      else if (pg_with_nat) /* If the process group has a non-cygwin process,
+			       it cannot be suspended correctly. Therefore,
+			       do not send SIGTSTP. */
+	goto not_a_sig;
+      else if (CCEQ (ti.c_cc[VSUSP], c))
+	sig = SIGTSTP;
+      else
+	goto not_a_sig;
+
+      termios_printf ("got interrupt %d, sending signal %d", c, sig);
+      if (!(ti.c_lflag & NOFLSH) && fh)
+	{
+	  fh->eat_readahead (-1);
+	  fh->discard_input ();
+	}
+      if (fh)
+	fh->release_input_mutex_if_necessary ();
+      ttyp->kill_pgrp (sig, pgid);
+      if (fh)
+	fh->acquire_input_mutex_if_necessary (mutex_timeout);
+      ti.c_lflag &= ~FLUSHO;
+      return signalled;
+    }
+not_a_sig:
+  if ((ti.c_lflag & ISIG) && need_discard_input)
+    {
+      if (!(ti.c_lflag & NOFLSH) && fh)
+	{
+	  fh->eat_readahead (-1);
+	  fh->discard_input ();
+	}
+      ti.c_lflag &= ~FLUSHO;
+      return not_signalled_but_done;
+    }
+  bool to_nat = !cyg_reader && pg_with_nat;
+  return to_nat ? not_signalled_with_nat_reader : not_signalled;
+}
+
+bool
+fhandler_termios::process_stop_start (char c, tty *ttyp)
+{
+  termios &ti = ttyp->ti;
+  if (ti.c_iflag & IXON)
+    {
+      if (CCEQ (ti.c_cc[VSTOP], c))
+	{
+	  ttyp->output_stopped = true;
+	  return true;
+	}
+      else if (CCEQ (ti.c_cc[VSTART], c))
+	{
+restart_output:
+	  ttyp->output_stopped = false;
+	  return true;
+	}
+      else if ((ti.c_iflag & IXANY) && ttyp->output_stopped)
+	goto restart_output;
+    }
+  if ((ti.c_lflag & ICANON) && (ti.c_lflag & IEXTEN)
+      && CCEQ (ti.c_cc[VDISCARD], c))
+    {
+      ti.c_lflag ^= FLUSHO;
+      return true;
+    }
+  return false;
+}
+
 line_edit_status
 fhandler_termios::line_edit (const char *rptr, size_t nread, termios& ti,
 			     ssize_t *bytes_read)
@@ -326,51 +546,25 @@ fhandler_termios::line_edit (const char *rptr, size_t nread, termios& ti,
 
       if (ti.c_iflag & ISTRIP)
 	c &= 0x7f;
-      if (ti.c_lflag & ISIG)
+      bool disable_eof_key = false;
+      switch (process_sigs (c, get_ttyp (), this))
 	{
-	  int sig;
-	  if (CCEQ (ti.c_cc[VINTR], c))
-	    sig = SIGINT;
-	  else if (CCEQ (ti.c_cc[VQUIT], c))
-	    sig = SIGQUIT;
-	  else if (CCEQ (ti.c_cc[VSUSP], c))
-	    sig = SIGTSTP;
-	  else
-	    goto not_a_sig;
-
-	  termios_printf ("got interrupt %d, sending signal %d", c, sig);
-	  if (!(ti.c_lflag & NOFLSH))
-	    {
-	      eat_readahead (-1);
-	      discard_input ();
-	    }
-	  release_input_mutex_if_necessary ();
-	  tc ()->kill_pgrp (sig);
-	  acquire_input_mutex_if_necessary (mutex_timeout);
-	  ti.c_lflag &= ~FLUSHO;
+	case signalled:
 	  sawsig = true;
-	  goto restart_output;
-	}
-    not_a_sig:
-      if (ti.c_iflag & IXON)
-	{
-	  if (CCEQ (ti.c_cc[VSTOP], c))
-	    {
-	      if (!tc ()->output_stopped)
-		tc ()->output_stopped = true;
-	      continue;
-	    }
-	  else if (CCEQ (ti.c_cc[VSTART], c))
-	    {
-    restart_output:
-	      tc ()->output_stopped = false;
-	      continue;
-	    }
-	  else if ((ti.c_iflag & IXANY) && tc ()->output_stopped)
-	    goto restart_output;
+	  fallthrough;
+	case not_signalled_but_done:
+	case done_with_debugger:
+	  get_ttyp ()->output_stopped = false;
+	  continue;
+	case not_signalled_with_nat_reader:
+	  disable_eof_key = true;
+	  break;
+	default: /* Not signalled */
+	  break;
 	}
+      if (process_stop_start (c, get_ttyp ()))
+	continue;
       /* Check for special chars */
-
       if (c == '\r')
 	{
 	  if (ti.c_iflag & IGNCR)
@@ -389,12 +583,6 @@ fhandler_termios::line_edit (const char *rptr, size_t nread, termios& ti,
 	    set_input_done (iscanon);
 	}
 
-      if (iscanon && ti.c_lflag & IEXTEN && CCEQ (ti.c_cc[VDISCARD], c))
-	{
-	  ti.c_lflag ^= FLUSHO;
-	  continue;
-	}
-
       if (!iscanon)
 	/* nothing */;
       else if (CCEQ (ti.c_cc[VERASE], c))
@@ -431,7 +619,7 @@ fhandler_termios::line_edit (const char *rptr, size_t nread, termios& ti,
 	    }
 	  continue;
 	}
-      else if (CCEQ (ti.c_cc[VEOF], c))
+      else if (CCEQ (ti.c_cc[VEOF], c) && !disable_eof_key)
 	{
 	  termios_printf ("EOF");
 	  accept_input ();
